#include <exception>
#include <map>
#include <vector>
#include <unistd.h>
#include <math.h>
#include <stdio.h>

#include "gme/Music_Emu.h"
#include "gme/KeyLog.h"
#include "nez.h"

static const int fd_audioout  = 1;
static const int fd_trackdump = 3;

#include "ch.hh"

static void SaveMelody(const std::vector<situation>& track, unsigned loop_begin)
{
    FILE* out = fdopen(fd_trackdump, "w");
    fprintf(out, "%u %u\n", track.size(), loop_begin);
    for(unsigned a=0; a<track.size(); ++a)
    {
        fprintf(out, "%u ", track[a].size());
        for(situation::const_iterator
            i = track[a].begin();
            i != track[a].end();
            ++i)
        {
            fprintf(out, "%u %f %f ", i->first, i->second.amplitude, i->second.hz);
        }
        fprintf(out, "\n");
    }
    fclose(out);
}

static situation st;
void KeyLogOn(unsigned channel, double hertz, double amplitude)
{
    if(amplitude < silence_level) KeyLogOff(channel);
    else
    {
        ch tmp = { hertz, amplitude };
        st[channel]=tmp;
    }
}
void KeyLogOff(unsigned channel)
{
    st.erase(channel);
}

static void DumpSituation()
{
    for(situation::const_iterator
        i = st.begin();
        i != st.end();
        ++i)
    {
        fprintf(stderr, "%u(%g %g)", i->first, i->second.amplitude, i->second.hz);
    }
    fprintf(stderr, "\n");
}

static bool IsSilent()
{
    return st.empty();
}

static const double Func(double n)
{
    double ret = sin(n);
    if(ret < -0.3) return ret;
    if(ret > 0.3) return ret;
    return 3*ret;//et;
}

static void AudioDump(const short* buf2, int size)
{
    if(fd_audioout < 0) return;
    return;
    
    short buf[size];
    static int angle=0;
    for(int a=0; a<size; a+=2)
    {
        double n=0;
        for(situation::const_iterator
            i = st.begin();
            i != st.end();
            ++i)
                n += i->second.amplitude *
                    Func(angle * M_PI * 2 * i->second.hz / sample_rate);
        buf[a+0] = buf2[a+0]; //(short)(n * 4000);
        buf[a+1] = buf2[a+1]; //(short)((buf2[a+0] - buf2[a+1])*0.45);
        ++angle;
    }
    write(fd_audioout, buf, size*2);
}

int main(int argc, char**argv)
{
    std::set_terminate (__gnu_cxx::__verbose_terminate_handler);

    char TmpFileName[] = "/tmp/gmenezXXXXXX";
    void* nezfile = 0;
    unsigned nezfilesize = NEZ_extract(argv[1], &nezfile);
    if(!nezfilesize) throw "No file to extract";
    int fd = mkstemp(TmpFileName);
    write(fd, nezfile, nezfilesize);
    close(fd);
    
    gme_type_t file_type;
    gme_identify_file(TmpFileName, &file_type);
    if(!file_type) throw "Unsupported music type";
    
    Music_Emu* emu = file_type->new_emu();
    emu->set_sample_rate(sample_rate);
    emu->load_file(TmpFileName);
    
    unlink(TmpFileName);

    int ntracks = emu->track_count();
    int trackno=argv[2]?atoi(argv[2]):0;
    if(trackno >= ntracks) return -1;
    
    //for(int trackno=0; trackno < ntracks; ++trackno)
    {
        emu->ignore_silence(true);
        emu->start_track(trackno);
        
        unsigned n_silent    = 0;
        std::vector<situation> track;
        
        const unsigned minimum_length_before_checking_loops = 60 * 60; // 1 minute
        const unsigned minimum_silence_before_automatic_end = 60 * 15; // 15 secs
        const unsigned minimum_length_before_any_rep_works = 60 * 600; // 10 minutes
        const unsigned minimum_autoaccept_rep_total_length = 60 * 40;
        // ^ 40 seconds of looping material = automatically accept
        const unsigned loop_checking_interval = 60*10;
        
        unsigned last_checked_loopendpos = 0;
        
        unsigned best_rep_length = 0;
        unsigned best_rep_pos    = 0;
        unsigned best_rep_count  = 0;
        
        //while(emu->tell() < 0.4*1000L)
        
        track.reserve(minimum_length_before_checking_loops);
        
        static double ticks_fraction = 0;
        for(;;)
        {
            ticks_fraction += 2*sample_rate/ticks_per_sec;
            long size = ((int)ticks_fraction) & ~1;
            ticks_fraction -= size;
        
            short audiobuf[sample_rate * 10];
            emu->play_until_nmi(audiobuf, size);
            
            //DumpSituation();
            AudioDump(audiobuf, size);
            
            if(IsSilent())
            {
                if(track.empty()) continue;
                ++n_silent;
                if(n_silent >= minimum_silence_before_automatic_end)
                {
                    track.resize(track.size() - n_silent);
                    fprintf(stderr, "silence termination @ %.2f\n", track.size()/60.0);
                    goto done_track_play;
                }
            }
            else
                n_silent=0;

            track.push_back(st);
            
            if(track.size() >= minimum_length_before_checking_loops
            && track.size() >= last_checked_loopendpos + loop_checking_interval)
            {
                /* Attempt to find loops */
                unsigned check_pos = std::max(last_checked_loopendpos, track.size() / 2);
                //fprintf(stderr, "loopcheck @ %.2f\n", check_pos/60.0);
                while(check_pos < track.size())
                {
                    const unsigned min_rep_length = std::max(60U * 3, best_rep_length);
                    const unsigned max_rep_length = check_pos/2;
                    for(unsigned rep_length = max_rep_length;
                                 rep_length >= min_rep_length;
                                 --rep_length)
                    {
                        const unsigned max_rep = check_pos / rep_length;
                        unsigned rep_count;
                        
                        if(max_rep < 2)
                            fprintf(stderr, "error, pos(%u)len(%u) max=%u\n",
                                check_pos,rep_length,max_rep);
                        
                        for(rep_count=1; rep_count < max_rep; ++rep_count)
                        {
                            for(unsigned r=0; r<rep_length; ++r)
                                if(track[check_pos-r] !=
                                track[check_pos-rep_count*rep_length-r])
                                {
                                    //if(r>60)
                                    //fprintf(stderr, "rep fail @ %u/%u\n", r,rep_length);
                                    goto not_rep;
                                }
                            continue; // Found a repetition!
                        not_rep: break;
                        }
                        if(rep_count == 1) continue; // no repetition
                        
                    Redo_subdivide:
                        /* Check if it can be subdivided */
                        const unsigned min_subdivided_size = 60U*3;
                        for(unsigned subdivide_count=2; ; ++subdivide_count)
                        {
                            unsigned subdivide_length = rep_length / subdivide_count;
                            if(subdivide_length < min_subdivided_size) break;
                            if(rep_length % subdivide_count) continue;
                            
                            for(unsigned s=0; s<subdivide_length; ++s)
                                for(unsigned p=1; p<subdivide_count; ++p)
                                    if(track[check_pos-s] !=
                                    track[check_pos-p*subdivide_length-s]) goto not_subrep;
                            /*
                            fprintf(stderr, "subdivided %u*%u by %u\n",
                                rep_length,rep_count,subdivide_count);
                            */
                            rep_length /= subdivide_count;
                            rep_count *= subdivide_count;
                            goto Redo_subdivide;
                        not_subrep:;
                        }
                        
                        // Found!
                        if(rep_length > best_rep_length
                        || (rep_length == best_rep_length && rep_count > best_rep_count))
                        {
                            /*
                            fprintf(stderr,
                                "Found a rep (len=%u, pos=%u, count=%u)\n",
                                    rep_length, check_pos, rep_count);
                            */
                            best_rep_length = rep_length;
                            best_rep_pos    = check_pos;
                            best_rep_count  = rep_count;
                            
                            unsigned rep_amount = rep_length * rep_count;
                            if(rep_amount >= minimum_autoaccept_rep_total_length)
                            {
                                goto done_track_play;
                            }
                            
                            // Break, because it doesn't make sense to test
                            // shorter reps (we're only interested in the longest).
                            break;
                        }
                    }
                    
                    ++check_pos;
                }
                last_checked_loopendpos = check_pos;
            }
            if(track.size() >= minimum_length_before_any_rep_works && best_rep_length)
            {
            done_track_play:
                //fprintf(stderr, "done_track_play.\n");
                break;
            }
        }
        
        unsigned loop_begin = track.size();
        if(best_rep_length > 0)
        {
            loop_begin = best_rep_pos - best_rep_length * best_rep_count;
            while(loop_begin > 0
               && track[loop_begin-1] == track[loop_begin-1 + best_rep_length])
            {
                --loop_begin;
            }
            track.resize(loop_begin + best_rep_length);
            best_rep_count = 1;
            
            fprintf(stderr, "Track reduced to %u frames intro + %u frames loop\n",
                loop_begin, best_rep_length);
        }
        
        SaveMelody(track, loop_begin);
    }
    
    delete emu;
}
